"""
This module contains functionality for accessing the realtime parts of
the PacketLogic system, such as system diagnostics, channel statistics
and dynamic netobject items.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> rt = pl.Realtime()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from thisversion import verify_ip, negotiate_flicka_with_timeout, deprecatedalias
from thisversion import PL
from thisversion import PLrcd
from thisversion import CommitID

import socket
import select
import time
import sys

from packetlogic2.exceptions import PLDPermissionError, PLDNotAvailable, PLDUnsupportedFeatureError, PLDServerError, PLDUnsupportedInFirmware

class KeyVault(object):
	"""
	Operations on the key vault used for encrypted property objects.

		>>> # Set key at index 3
		>>> rt.keyvault.set(3, "abc123")
		>>> # Add an property object with encryption
		>>> obj = rs.object_add("/PropertyObject/Sensitive")
		>>> o.attributes['encrypted'] = "3"
	"""
	def __init__(self, plrcd):
		self.plrcd = plrcd

	def set(self, index, key):
		"""
		Set a key in the key vault to the specified string.

		@type index: C{int}
		@param index: Index of the key to remove.

		@type key: C{str}
		@param key: The new encryption key.

		@attention: key must not contain any '\0' character.
		"""
		cid = PLrcd.doKeyvaultSet(self.plrcd.io, index, key)
		rcid = self.plrcd._do(PLrcd.handleKeyvault, PLrcd.PROTOCOL_SERVER_KEYVAULT_SET)
		assert cid == rcid

	def reset(self, index):
		"""
		Remove a key from the key vault.

		@type index: C{int}
		@param index: Index of the key to remove.
		"""
		cid = PLrcd.doKeyvaultReset(self.plrcd.io, index)
		rcid = self.plrcd._do(PLrcd.handleKeyvault, PLrcd.PROTOCOL_SERVER_KEYVAULT_RESET)
		assert cid == rcid

class Dyn(object):
	"""
	Operations on dynamic netobject items.

	@cvar SUBSCRIBER_NAME: Constant for the "subscriber name" key in property lists.
	@cvar IS_SUBSCRIBER: Constant for the "is subscriber" key in property lists.
	@cvar LINKSPEED_IN: Constant for the "inbound linkspeed" key in property lists.
	@cvar LINKSPEED_OUT: Constant for the "outbound linkspeed" key in property lists.
	@cvar NATCFG: Constant for the "nat config" key in property lists.
	"""
	SUBSCRIBER_NAME = PLrcd.DYNITEM_PROPERTY_SUBSCRIBER_NAME
	IS_SUBSCRIBER = PLrcd.DYNITEM_PROPERTY_IS_SUBSCRIBER
	LINKSPEED_IN = PLrcd.DYNITEM_PROPERTY_LINKSPEED_IN
	LINKSPEED_OUT = PLrcd.DYNITEM_PROPERTY_LINKSPEED_OUT
	NATCFG = PLrcd.DYNITEM_PROPERTY_NATCFG

	def __init__(self, plrcd):
		self.plrcd = plrcd

	def list(self, netobject=None, ip=None, recurse=False, properties=[]):
		"""
		List dynamic netobject items.

			>>> rt.dyn.list(netobject=300, recurse=True, properties=[rt.dyn.SUBSCRIBER_NAME])
			[('10.0.0.2',
			  [(339L, [{1: 'User0001'}]),
			   (337L, [{1: 'User0001'}])]),
			 ('10.0.0.3',
			  [(339L, [{1: 'User0002'}]),
			   (335L, [{1: 'User0002'}])])]

		@type netobject: C{int}
		@param netobject: Only list items in specified netobject (or any netobject under it if recurse parameter is True)

		@type ip: C{str}
		@param ip: Only list items for specified ip/mask address (or any ip under it if recurse parameter is True)

		@type recurse: C{bool}
		@param recurse: Also list items below specified netobject/ip

		@type properties: C{list}
		@param properties: Properties to include in result. Should be a list containing zero or more of  SUBSCRIBER_NAME, IS_SUBSCRIBER, LINKSPEED_IN, LINKSPEED_OUT, NATCFG

		@rtype: C{list} of C{tuple}
		@returns: A list of dynitems grouped by ip address, e.g [(IP1, [(NO1, [PROPS..]), (NO2, [PROPS])]), (IP2, [(NO1, [PROPS..]), (NO2, [PROPS])])]

		"""

		if netobject and ip:
			raise ValueError("Can't specify both netobject and ip")

		if recurse and (not (netobject or ip)):
			raise ValueError("Can't specify recurse without netobject or ip")

		if netobject and not recurse:
			selt, selv = PLrcd.PROTOCOL_CLIENT_DYNNETOBJ_LIST_SELECTOR_NO, netobject
		elif netobject and recurse:
			selt, selv = PLrcd.PROTOCOL_CLIENT_DYNNETOBJ_LIST_SELECTOR_NO_RECURSE, netobject
		elif ip and not recurse:
			selt, selv = PLrcd.PROTOCOL_CLIENT_DYNNETOBJ_LIST_SELECTOR_IP, ip
		elif ip and recurse:
			selt, selv = PLrcd.PROTOCOL_CLIENT_DYNNETOBJ_LIST_SELECTOR_IP_RECURSE, ip
		else:
			selt, selv = PLrcd.PROTOCOL_CLIENT_DYNNETOBJ_LIST_SELECTOR_ALL, None

		cid = PLrcd.dynList(self.plrcd.io, selt, selv, properties)

		lst = []
		def collector(msg):
			rcid, ip, items = PLrcd.handleDynItem(msg)
			assert cid == rcid
			oitems = [(no, dict(props)) for no, props in items]
			lst.append((ip, oitems))

		self.plrcd._handler_add(PLrcd.PROTOCOL_SERVER_DYNNETOBJ_LIST_ITEM, collector)
		rcid, r, msg = self.plrcd._do(PLrcd.handleDynResponse, PLrcd.PROTOCOL_SERVER_DYNNETOBJ_RESPONSE)
		assert cid == rcid
		self.plrcd._handler_remove(PLrcd.PROTOCOL_SERVER_DYNNETOBJ_LIST_ITEM)
		if r != 0:
			raise PLDServerError(msg)

		return lst

	def add(self, no, ip, subscriber="", is_subscriber=False, linkspeed_in=0, linkspeed_out=0, natcfg=""):
		"""
		Add a dynamic netobject item

		Please note that netobject will not be verified with regard to
		performance and flexibility when adding to a uncommitted
		netobject.

			>>> rt.dyn.add(31682, '127.0.0.1')

		@type id: C{int}
		@param id: the id number of the netobject to add the item to.

		@type ip: C{str}
		@param ip: the IP address to add.

		@type subscriber: C{str}
		@param subscriber: Subscriber name - or empty string to not create subscriber.

		@type is_subscriber: C{bool}
		@param is_subscriber: is subscriber property - used for subscriber count in statistics.

		@type linkspeed_in: C{int}
		@param linkspeed_in: linkspeed_in property - used in statistics.

		@type linkspeed_out: C{int}
		@param linkspeed_out: linkspeed_out property - used in statistics.

		@type natcfg: C{str}
		@param natcfg: natcfg property - used to config subscribers NAT. Requires v14.0 firmware or newer.

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDServerError: If server reported an error
		@raises ValueError: If an invalid IP address is used
		"""
		if not self.plrcd.dyn_add(no, ip, subscriber, is_subscriber, linkspeed_in, linkspeed_out, natcfg):
			raise PLDServerError(self.plrcd._last_dyn_msg)

	def remove(self, no, ip):
		"""
		Remove a dynamic netobject item

			>>> rt.dyn.remove(31682, '127.0.0.1')

		@type no: C{int}
		@param no: the id number of the netobject to remove the item from.

		@type ip: C{str}
		@param ip: the ip number to remove.

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDServerError: If server reported an error
		"""
		if not self.plrcd.dyn_remove(no, ip):
			raise PLDServerError(self.plrcd._last_dyn_msg)

	def set_begin(self, baseno):
		"""
		Starts a session of following set() operations under a base netobject.

		When zero or more set() have been performed and set_end() is called,
		all dynamic netobject items under the baseno NetObject that were not
		set by any set() will be removed.

		Performing a series of set() operations between a pair of
		set_begin() and set_end() will make sure that all dynamic netobject
		items under the baseno NetObject will be exactly defined by the of
		set() operations.

		Note that a new set_begin() can not be issued until the
		corresponding closing set_end() has been performed.

		Also note that all following set() operations must have the same
		baseno NetObject id as the one provided in set_begin().

		@type baseno: C{int}
		@param baseno: the id number of the netobject to operate under.

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.
		@raises PLDServerError: If server reported an error
		"""
		cid = PLrcd.dynSetBegin(self.plrcd.io, baseno)
		rcid, r, msg = self.plrcd._do(PLrcd.handleDynResponse, PLrcd.PROTOCOL_SERVER_DYNNETOBJ_RESPONSE)
		assert cid == rcid
		if r != 0:
			raise PLDServerError(msg)

	def set(self, baseno, ip, items):
		"""
		Set multiple dynamic netobject items under a base netobject.

			>>> rt.dyn.set(300, "10.0.0.4", [(339, {rt.dyn.SUBSCRIBER_NAME: 'User0003', rt.dyn.IS_SUBSCRIBER: True}),
			...                              (337, {rt.dyn.SUBSCRIBER_NAME:'User0003', rt.dyn.IS_SUBSCRIBER: True})])
			>>> rt.dyn.set(300, "10.0.0.3", [])

		This function replaces all dynamic netobject items
		that exists for the specified ip address in the
		netobject tree under the specified base netobject id.

		@type baseno: C{int}
		@param baseno: the id number of the netobject to operate under.

		@type ip: C{str}
		@param ip: the ip number to set items for.

		@type items: C{sequence} of C{tuples}
		@param items: the items to add, each item should be a
			      tuple of (netobjectid, properties),
			      where properties is a dict which may
			      contain zero or more of the keys
			      rt.dyn.SUBSCRIBER_NAME,
			      rt.dyn.IS_SUBSCRIBER,
			      rt.dyn.LINKSPEED_IN,
			      rt.dyn.LINKSPEED_OUT and
			      rt.dyn.NATCFG

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.
		@raises PLDServerError: If server reported an error
		"""
		cid = PLrcd.dynSetNew(self.plrcd.io, baseno, ip, items)
		rcid, r, msg = self.plrcd._do(PLrcd.handleDynResponse, PLrcd.PROTOCOL_SERVER_DYNNETOBJ_RESPONSE)
		assert cid == rcid
		if r != 0:
			raise PLDServerError(msg)

	def set_end(self, baseno, abort=False):
		"""
		Ends a session of set() operations, started by set_begin(), under a base netobject.

		When zero or more set() have been performed and set_end() is called
		all dynamic netobject items under the baseno  NetObject that were
		not set by any set() operation will be removed.

		Performing a series of set() operations between a pair of
		set_begin() and set_end() will make sure that all dynamic NetObject
		items under the baseno NetObject will be exactly defined by the
		series of dyn_set() operations.

		@type baseno: C{int}
		@param baseno: Id number of the netobject to operate under.

		@type abort: C{bool}
		@param abort: Set to True to abort the session without
		              removing items belonging to ip addresses
		              that did not receive a call to set() in
		              the session.

		@rtype: C{int}
		@returns: Number of removed dynamic netobject items that were not set by any set().

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.
		@raises PLDServerError: If server reported an error
		"""
		cid = PLrcd.dynSetEnd(self.plrcd.io, baseno, abort)
		rcid, r, msg = self.plrcd._do(PLrcd.handleDynResponse, PLrcd.PROTOCOL_SERVER_DYNNETOBJ_RESPONSE)
		assert cid == rcid
		if r != 0:
			raise PLDServerError(msg)

	def remove_orphans(self):
		"""
		Removes all orphan dynamic netobject items. An orphan dynamic
		netobject item is a dynamic netobject item created in a netobject
		that no longer exists. For example:

			>>> rs=pl.Ruleset()
			>>> rt=pl.Realtime()
			>>> o=rs.object_add("/NetObjects/example")
			>>> rs.commit()
			>>> # Wait for plrcd to read up the commit

			>>> rt.dyn.add(o.id, '127.0.0.1', "subscriber")
			True
			>>> rs.object_remove(o)
			>>> rs.commit()
			>>> # Wait for plrcd to read up the commit

			>>> rt.dyn.remove_orphans()
			1L

		Note that created/removed object will not appear/disappear until
		the change has been committed. Be warned that the following might
		remove wanted dynamic netobject items:

			>>> rs=pl.Ruleset()
			>>> rt=pl.Realtime()
			>>> o=rs.object_add("/NetObjects/example2")
			>>> rt.dyn.add(o.id, '127.0.0.2', "subscriber2")
			True
			>>> rt.dyn.remove_orphans()
			1L

		In this example the newly added dynamic netobject item
		'127.0.0.2' will be removed by remove_orphan() as the
		parent netobject was not yet committed.

		@rtype: C{int}
		@returns: Number of removed orphans dynamic netobject items

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDServerError: If server reported an error
		"""
		class changes:
			def __init__(self):
				self.removed = None
			def handler(self, msg):
				rcid, _, self.removed = PLrcd.handleDynChanges(msg)
				assert cid == rcid

		cid = PLrcd.dynRemoveOrphans(self.plrcd.io)
		mychanges = changes()
		self.plrcd._handler_add(PLrcd.PROTOCOL_SERVER_DYNNETOBJ_CHANGES, mychanges.handler)
		rcid, r, msg = self.plrcd._do(PLrcd.handleDynResponse, PLrcd.PROTOCOL_SERVER_DYNNETOBJ_RESPONSE)
		assert cid == rcid
		self.plrcd._handler_remove(PLrcd.PROTOCOL_SERVER_DYNNETOBJ_CHANGES)
		if r != 0:
			raise PLDServerError(msg)

		assert mychanges.removed is not None #Assert if we have not received a CHANGES msg
		return mychanges.removed

class PLRCD(object):
	"""
	Class for connecting and communicating with plrcd,
	the realtime part of PacketLogic.

	When using the multiversion API you will access this class via the
	L{PLConnection.Realtime() <packetlogic2.PLConnection.Realtime>}
	method.

	All functions that communicate with the PacketLogic Realtime
	system might raise errors common to all or many of the
	functions.Here is a list of such errors, that might not be
	listed on each method:

	@raises ValueError: As the superclass of an internal BadMessageError,
	if a malformed message is recieved on the PacketLogic connection.
	That normally will never happen though.

	@raises Realtime.PLDPermissionError: If the user lacks permission
	        for the requested action.

	@raises IOError: If a generic I/O error occurs while reading or
	        writing on the PacketLogic Realtime connection.

	@group Dynamic NetObject items: dyn_add,dyn_list,dyn_list_full,dyn_list_mask,dyn_list_no,dyn_remove
	@group Firewall Log: fwlog_clear,fwlog_query_offsets,fwlog_query_rule,fwlog_query_time_range
	@group Volume Based Shaping: vbs_list,vbs_query,vbs_reset,vbs_set
	@group Internal: dump_bgptable,dump_drdlstats,inject_packet,reload_drdl
	@group Realtime Data: add_channels_callback,add_netobj_callback,add_service_callback,add_shapingcnt_callback,add_shapingobject_callback,add_sysdiag_callback,add_stats_callback,add_update_done_callback,get_channels_data,get_netobj_data,get_services_data,get_sysdiag_data,get_stats_data,shapingcnt_request_all,stop_updating,update_done_callbacks,update_forever
	
	"""
	
	SUPPORTS_TIMEOUT = True # Actually it is pldb.Ruleset that is checked for this flag, but it seems apropriate to set this too.
	
	def __init__(self, host, user, password, port=42005, timeout=60.0):
		"""
		Initialize and connect to packetlogicd.

		@type host: C{str}
		@param host: A hostname or IP address for the PacketLogic
		             system you want to connect to.

		@type user: C{str}
		@param user: The username you want to authenticate
		             with.  The system share users with the
		             client. You can create the user you want
		             to use with the client.

		@type password: C{str}
		@param password: The password to authenticate with.

		@type port: C{int}
		@param port: The default port is 42001 and normally
		             you don't have to change that.

		@type timeout: C{float} or C{None}
		@param timeout: The connection timeout in seconds. Default None.
			Set to None to use blocking mode during connect, with no specified timeout.
			(A system dependent timeout may or may not still occur after some time.)
			New in version v11.3beta3
		       
		@raises RuntimeError: If authorization fails
		@raises RuntimeError: If server and API version does not match
		@raises IOError: B{"Operation timed out"}: If operation times out without response.
		@raises IOError: If an error occures on the socket when attempting to connect.
		
		"""
		self.host = host
		self.port = port
		self.user = user
		self.timeout = timeout
		#self.timeout = timeout or 60.0 # FIXME Should we set this to the timeout argument too?
		self.password = password
		self.authed = False
		self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.s.setblocking(False)
		self.s.settimeout(timeout)
		try:
			self.s.connect((host, port))
		except socket.error:
			typ, val, tb = sys.exc_info()
			raise IOError, val, tb
		self.io = PL.PLIO(self.s.fileno())
		self.handlers = {}
		self.shapingcnt_callbacks = []
		self.update_done_callbacks = []
		self.commit_bcast = None

		self._last_dyn_msg = "No error"

		negotiate_flicka_with_timeout(self.s, self.io, timeout)

		self.handlers[PLrcd.PROTOCOL_SERVER_VERSION] = self._handle_vers
		self.handlers[PLrcd.PROTOCOL_SERVER_AUTHENTICATION_OK] = self._handle_auth
		self.handlers[PLrcd.PROTOCOL_SERVER_EPERM] = self._handle_eperm
		self.handlers[PLrcd.PROTOCOL_SERVER_NOT_AVAILABLE] = self._handle_eagain
		self.handlers[PLrcd.PROTOCOL_SERVER_UPDATE_DONE] = self._handle_update_done
		self.handlers[PLrcd.PROTOCOL_SERVER_SHAPINGCOUNTER_ENTRY] = self._handle_shapingcnt
		self.handlers[PLrcd.PROTOCOL_SERVER_COMMIT_BCAST] = self._handle_commit_bcast

		PLrcd.doAuth(self.io, user, password)

		while not self.authed:
			if not self._process(self.timeout):
				raise IOError("Operation timed out")

	def _handler_add(self, cmd, handler):
		self.handlers[cmd] = handler

	def _handler_remove(self, cmd):
		if self.handlers.has_key(cmd):
			del self.handlers[cmd]

	def _handle_auth(self, m):
		if PLrcd.handleAuth(m) == 0:
			raise RuntimeError("Authorization failed")
		self.authed = True

	def _handle_vers(self, m):
		ver, instanceid, systemid, distversion = PLrcd.handleVersion(m)
		if PLrcd.PROTOCOL_VERSION == 0:
			import warnings
			warnings.warn("PythonAPI built without protocol version checking. Use at own risk")
		if ver == 0:
			import warnings
			warnings.warn("server built without protocol version checking. Use at own risk")

		if ((PLrcd.PROTOCOL_VERSION != 0) and
		    (ver != 0) and
		    (ver & ~0x7f) != (PLrcd.PROTOCOL_VERSION & ~0x7f)):
			raise RuntimeError("Bad server version (got %d expected %d)" % (ver, PLrcd.PROTOCOL_VERSION))
		self.protocolVersion = ver
		self.instanceid = instanceid
		self.systemid = systemid
		self.distversion = distversion

	def _handle_eperm(self, m):
		cmd = PLrcd.handleEperm(m)
		raise PLDPermissionError("Permission denied executing command (%x)" % cmd)

	def _handle_eagain(self, m):
		raise PLDNotAvailable("Data not available: '%s'" % PLrcd.handleNotAvail(m))

	def _fullflush(self):
		"""Make sure any outstanding messages are sent and
		   handled by the server by sending a ping request and
		   waiting for the response, which makes sure that all
		   messages sent before this are handled by server"""
		PLrcd.ping(self.io)
		self._do(None, PLrcd.PROTOCOL_SERVER_PONG)

	def _flush(self):
		while self.io.wantWrite():
			outp = [self.s]

			(i, o, e) = select.select([], outp, outp, self.timeout)

			if len(i) + len(o) + len(e) == 0:
				raise IOError("Operation timed out")

			if len(e) > 0:
				raise IOError("Select returned error")
			
			if len(o) > 0:
				if not self.io.doWrite():
					raise IOError("Error writing to socket")

	def _handle_update_done(self, m):
		for c in self.update_done_callbacks:
			c()

	def _handle_shapingcnt(self, m):
		r = PLrcd.shapingcounterHandleEntry(m)
		for f in self.shapingcnt_callbacks:
			f(*r)

	def _handle_commit_bcast(self, m):
		self.commit_bcast = CommitID(*PLrcd.handleCommitBcast(m));

	def _fill_socketlist(self, inl, outl):
		inl.append(self.s)
		if self.io.wantWrite():
			outl.append(self.s)

	def _handle_socket_events(self, i, o, e):
		if self.s in e:
			""" Huston! We have a problem! """
			raise IOError("Select returned error")

		if self.s in o:
			if not self.io.doWrite():
				raise IOError("Error writing to socket")

		if self.s in i:
			if not self.io.doRead():
				raise IOError("Error reading from socket")
		self._handle_msgs()

	def _handle_msgs(self):
		m = self.io.readMsg()
		while m:
			cmd = m.cmd()
			if self.handlers.has_key(cmd):
				if self.handlers[cmd](m):
					return True
			m = self.io.readMsg()
		return False


	def _process(self, timeout):
		"""
		This is an internal function that will process events as they
		occur on the socket
		"""

		if (self._handle_msgs()):
			return True

		inp = []
		outp = []
		self._fill_socketlist(inp, outp)

		(i, o, e) = select.select(inp, outp, inp, timeout)

		if len(i) + len(o) + len(e) == 0: # timeout
			return False

		self._handle_socket_events(i, o, e)
		return True

	def _request_update(self):
		PLrcd.requestUpdate(self.io)


	def _do(self, func, cmd):
		class handler:
			def __init__(self, pld):
				self.pld = pld
				self.ret = None
			def __call__(self, msg):
				del self.pld.handlers[cmd]
				if func:
					self.ret = func(msg)
				return True

		h = handler(self)
		self._handler_add(cmd, h)

		while self.handlers.has_key(cmd):
			if not self._process(self.timeout):
				raise IOError("Operation timed out")

		return h.ret

	def close (self):
		"""
		Immediately disconnect from packetlogic system.

		After this method has been called this object becomes
		useless.

		It may be used to force a disconnect when the garbage
		collector can't be trusted to do a timely disconnect.
		"""
		class Closed(object):
			def __getattr__(self, n):
				raise IOError("Closed")
		self.s.close()
		self.__dict__.clear()
		self.__class__ = Closed

	def dyn_add(self, id, ip, subscriber="", is_subscriber=False, linkspeed_in=0, linkspeed_out=0, natcfg=""):
		"""
		Add a dynamic netobject item

		Please note that netobject will not be verified with regard to
		performance and flexibility when adding to a uncommitted
		netobject.

			>>> rt.dyn_add(31682, '127.0.0.1')
			True

		@type id: C{int}
		@param id: the id number of the netobject to add the item to.

		@type ip: C{str}
		@param ip: the IP address to add.

		@type subscriber: C{str}
		@param subscriber: Subscriber name - or empty string to not create subscriber. Requires v12.0 firmware or newer.

		@type is_subscriber: C{bool}
		@param is_subscriber: is subscriber property - used for subscriber count in statistics. Requires v12.4 firmware or newer.

		@type linkspeed_in: C{int}
		@param linkspeed_in: linkspeed_in property - used in statistics. Requires v12.4 firmware or newer.

		@type linkspeed_out: C{int}
		@param linkspeed_out: linkspeed_out property - used in statistics. Requires v12.4 firmware or newer.

		@type natcfg: C{str}
		@param natcfg: natcfg property - used to config subscribers NAT. Requires v14.0 firmware or newer.

		@rtype: C{boolean}
		@returns: True if add was successful

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If subscriber argument is used and firmware doesn't support it.
		@raises ValueError: If an invalid IP address is used

		@attention: subscriber parameter requires v12.0 firmware or newer.
		@attention: is_subscriber parameter requires v12.4 firmware or newer.
		@attention: linkspeed_in parameter requires v12.4 firmware or newer.
		@attention: linkspeed_out parameter requires v12.4 firmware or newer.
		@attention: natcfg parameter requires v14.0 firmware or newer.
		"""
		cid = PLrcd.dynAdd(self.io, id, ip, subscriber, is_subscriber, linkspeed_in, linkspeed_out, natcfg)
		rcid, r, self._last_dyn_msg = self._do(PLrcd.handleDynResponse, PLrcd.PROTOCOL_SERVER_DYNNETOBJ_RESPONSE)
		assert cid == rcid
		if r == 0:
			return True
		return False


	def dyn_remove(self, id, ip):
		"""
		Remove a dynamic netobject item

			>>> rt.dyn_remove(31682, '127.0.0.1')
			True

		@type id: C{int}
		@param id: the id number of the netobject to remove the item from.

		@type ip: C{str}
		@param ip: the ip number to remove.

		@rtype: C{boolean}
		@returns: True if remove was successful

		@raises PLDPermissionError: If the user lacks permission for requested action
		"""
		cid = PLrcd.dynRemove(self.io, id, ip)
		rcid, r, self._last_dyn_msg = self._do(PLrcd.handleDynResponse, PLrcd.PROTOCOL_SERVER_DYNNETOBJ_RESPONSE)
		assert cid == rcid
		if r == 0:
			return True
		return False

	def dyn_set(self, baseno, ip, items):
		"""
		Set multiple dynamic netobject items under a base netobject.

		This function replaces all dynamic netobject items
		that exists for the specified ip address in the
		netobject tree under the specified base netobject id.

		@type baseno: C{int}
		@param baseno: the id number of the netobject to operate under.

		@type ip: C{str}
		@param ip: the ip number to set items for.

		@type items: C{sequence} of C{tuples}
		@param items: the items to add, each item should be a
		              tuple of (netobjectid, subscribername)
		              or (netobject, subscribername,
		              properties). Where properties is a dict
		              which may contain the keys
		              is_subscriber, linkspeed_in and
		              linkspeed_out

		@rtype: C{boolean}
		@returns: True if call was successful

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.4 firmware or newer.
		"""
		cid = PLrcd.dynSet(self.io, baseno, ip, items)
		rcid, r, self._last_dyn_msg = self._do(PLrcd.handleDynResponse, PLrcd.PROTOCOL_SERVER_DYNNETOBJ_RESPONSE)
		assert cid == rcid
		if r == 0:
			return True
		return False

	def dyn_strerror(self):
		"""
		Return last error message from dyn_add/dyn_remove/dyn_set

			>>> rt.dyn_add(19, '10.0.0.1')
			True
			>>> rt.dyn_add(19, '10.0.0.1')
			False
			>>> rt.dyn_strerror()
			'Item already exists'

		@rtype: C{str}
		@returns: Last error message

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.4 firmware or newer.
		"""
		return self._last_dyn_msg

	class _ListCollector:
		def __init__(self, func):
			self.func = func
			self.lst = []

		def __call__(self, msg):
			self.func(msg, self.lst)

	_dyn_doc = """
		Get a handle for dynamic netobject operations. See L{Dyn} object for details.
"""
	dyn = property(lambda self: Dyn(self), doc=_dyn_doc)

	def dyn_list(self):
		"""
		List all dynamic netobject items

			>>> rt.dyn_list()
			[(31589, '127.0.0.1'),
			 (31576, '192.168.1.1'),
			 (31589, '192.168.0.131'),
			 (31688, '192.168.0.131'),
			 (31577, '192.168.1.1'),
			 (31682, '127.0.0.1')]

		@rtype: C{list} of C{tuple}
		@return: a list of tuples containing the netobject id and the item.

		@raises PLDNotAvailable: If data is not ready yet, and you should retry this call later. (This is not an error.)
		@raises PLDPermissionError: If the user lacks permission for requested action
		"""
		cid = PLrcd.dynList(self.io,
				    PLrcd.PROTOCOL_CLIENT_DYNNETOBJ_LIST_SELECTOR_ALL, None,
				    [])
		return self._dyn_list_handle(cid, False)

	def dyn_list_full(self):
		"""
		List all dynamic netobject items including subscriber names

			>>> rt.dyn_list()
			[(31589, '127.0.0.1', 'User 1'),
			 (31576, '192.168.1.1', 'User 123'),
			 (31589, '192.168.0.131', 'User 124'),
			 (31688, '192.168.0.131', 'User 124'),
			 (31577, '192.168.1.1', 'User 123'),
			 (31682, '127.0.0.1', 'User 1')]

		@rtype: C{list} of C{tuple}
		@return: a list of tuples containing the netobject id, the item and subscriber name.

		@raises PLDNotAvailable: If data is not ready yet, and you should retry this call later. (This is not an error.)
		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.0 firmware or newer.
		"""
		cid = PLrcd.dynList(self.io,
				    PLrcd.PROTOCOL_CLIENT_DYNNETOBJ_LIST_SELECTOR_ALL, None,
				    [PLrcd.DYNITEM_PROPERTY_SUBSCRIBER_NAME])
		return self._dyn_list_handle(cid, True)

	def _dyn_list_handle(self, cid, includesub):
		lst = []
		def collector(msg):
			rcid, ip, items = PLrcd.handleDynItem(msg)
			assert cid == rcid
			for no, attrs in items:
				if includesub:
					attrs = dict(attrs)
					itm = (no, ip, attrs.get(PLrcd.DYNITEM_PROPERTY_SUBSCRIBER_NAME,""))
				else:
					itm = (no, ip)
				lst.append(itm)

		self._handler_add(PLrcd.PROTOCOL_SERVER_DYNNETOBJ_LIST_ITEM, collector)
		rcid, r, errormsg = self._do(PLrcd.handleDynResponse, PLrcd.PROTOCOL_SERVER_DYNNETOBJ_RESPONSE)
		assert cid == rcid
		if r != 0:
			raise RuntimeError (errormsg)
		del self.handlers[PLrcd.PROTOCOL_SERVER_DYNNETOBJ_LIST_ITEM]

		return lst

	def dyn_list_no(self, id):
		"""
		List dynamic netobject items matching the specified netobject id

		This command is NOT supported on system with PacketLogic version v11.359
		or earlier. Using it will raise a timeout exception after about a minute.

			>>> rt.dyn_list_no(31576)
			[(31576, '192.168.1.1'),
			 (31576, '192.168.0.131'),
			 (31576, '192.168.2.13')]

		@type id: C{int}
		@param id: the id number of the netobject to match against.

		@rtype: C{list} of C{tuple}
		@return: a list of tuples containing the netobject id and the item.

		@raises PLDNotAvailable: If data is not ready yet, and you should retry this call later. (This is not an error.)
		@raises PLDPermissionError: If the user lacks permission for requested action

		"""
		cid = PLrcd.dynList(self.io,
				    PLrcd.PROTOCOL_CLIENT_DYNNETOBJ_LIST_SELECTOR_NO, id,
				    [])
		return self._dyn_list_handle(cid, False)

	def dyn_list_no_recurse(self, id):
		"""
		List dynamic netobject items matching the specified netobject id or an netobject under it.

			>>> rt.dyn_list_no_recurse(31576)
			[(31576, '192.168.1.1'),
			 (31577, '192.168.0.131'),
			 (31576, '192.168.2.13')]

		@type id: C{int}
		@param id: the id number of the parent netobject to match against.

		@rtype: C{list} of C{tuple}
		@return: a list of tuples containing the netobject id and the item.

		@raises PLDNotAvailable: If data is not ready yet, and you should retry this call later. (This is not an error.)
		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v13.1 firmware or newer.
		"""
		cid = PLrcd.dynList(self.io,
				    PLrcd.PROTOCOL_CLIENT_DYNNETOBJ_LIST_SELECTOR_NO_RECURSE, id,
				    [])
		return self._dyn_list_handle(cid, False)

	def dyn_list_mask(self, ip, mask):
		"""
		List all dynamic netobject items matching the specified ipaddress and netmask

		This command is NOT supported on system with PacketLogic version v11.359
		or earlier. Using it will raise a timeout exception after about a minute.

			>>> rt.dyn_list_mask('192.168.0.0', '255.255.252.0')
			[(31576, '192.168.1.1'),
			 (31589, '192.168.0.131'),
			 (31688, '192.168.0.131'),
			 (31577, '192.168.1.1'),
			 (31682, '192.168.2.33'),
			 (31697, '192.168.3.142')]

		@type ip: C{str}
		@param ip: the ip number to use for matching.

		@type mask: C{str}
		@param mask: the netmask to use for matching.

		@rtype: C{list} of C{tuple}
		@return: a list of tuples containing the netobject id and the item.

		@raises PLDNotAvailable: If data is not ready yet, and you should retry this call later. (This is not an error.)
		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDServerError: If either argument is of incorrect format.

		"""
		plen = {'0.0.0.0': 0,
			'128.0.0.0': 1,
			'192.0.0.0': 2,
			'224.0.0.0': 3,
			'240.0.0.0': 4,
			'248.0.0.0': 5,
			'252.0.0.0': 6,
			'254.0.0.0': 7,
			'255.0.0.0': 8,
			'255.128.0.0': 9,
			'255.192.0.0': 10,
			'255.224.0.0': 11,
			'255.240.0.0': 12,
			'255.248.0.0': 13,
			'255.252.0.0': 14,
			'255.254.0.0': 15,
			'255.255.0.0': 16,
			'255.255.128.0': 17,
			'255.255.192.0': 18,
			'255.255.224.0': 19,
			'255.255.240.0': 20,
			'255.255.248.0': 21,
			'255.255.252.0': 22,
			'255.255.254.0': 23,
			'255.255.255.0': 24,
			'255.255.255.128': 25,
			'255.255.255.192': 26,
			'255.255.255.224': 27,
			'255.255.255.240': 28,
			'255.255.255.248': 29,
			'255.255.255.252': 30,
			'255.255.255.254': 31,
			'255.255.255.255': 32}[mask]

		cid = PLrcd.dynList(self.io,
				    PLrcd.PROTOCOL_CLIENT_DYNNETOBJ_LIST_SELECTOR_IP_RECURSE, "%s/%s" % (ip, plen),
				    [])
		return self._dyn_list_handle(cid, False)

	def vbs_reset(self, machine, soid, extra=0):
		"""
		Reset vbs accounting data for a vbs object.

		@type machine: C{str}
		@param machine: MachineID of the PacketLogic you want to query.

		@type soid: C{int}
		@param soid: ID of VBS object

		@type extra: C{int}
		@param extra: Extra key; netobject id in case of split
		              by netobject, or ip number in case of
		              split by local host (encoded as an
		              integer). See the return value of
		              L{vbs_query} for details.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function isn't supported in v12.0 and v12.1 firmwares.
		"""
		PLrcd.vbsReset(self.io, machine, soid, extra)
		self._fullflush()


	def vbs_set(self, invalue, outvalue, machine, soid, extra=0):
		"""
		Set transfer to the specified value. Can only be used
		on object without sliding window (duration=1).

		@type invalue: C{int}
		@param invalue: The new value of inbound transfer (in KiB)

		@type outvalue: C{int}
		@param outvalue: The new value of outbound transfer (in KiB)

		@type machine: C{str}
		@param machine: MachineID of the PacketLogic you want to query.

		@type soid: C{int}
		@param soid: ID of VBS object

		@type extra: C{int}
		@param extra: Extra key; netobject id in case of split
		              by netobject, or ip number in case of
		              split by local host (encoded as an
		              integer). See the return value of
		              L{vbs_query} for details.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function does not exist in v12.0 firmware or newer.
		"""
		PLrcd.vbsReset(self.io, machine, soid, extra, invalue, outvalue)
		self._fullflush()

	def vbs_list(self):
		"""
		List available vbs objects

			>>> rt.vbs_list()
			[('00E081600B81', 24, 'VBS low', 1),
			 ('00E081600B81', 74, 'VBS medium', 1),
			 ('00E081600B81', 26, 'VBS max', 1)]

		@rtype: C{list} of C{tuple}
		@returns: A list of available VBS objects in a tuple (machineid,
		          id of vbs object, name, split type)
			  
			  split type is one of SPLIT_NONE, SPLIT_LOCALHOST,
			  SPLIT_HOST_NETOBJECT, SPLIT_SERVER_NETOBJECT,
			  SPLIT_CLIENT_NETOBJECT, SPLIT_CONNECTION

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function isn't supported in v12.0 and v12.1 firmwares.
		"""
		PLrcd.vbsList(self.io)
		return self._do(PLrcd.handleVbsList, PLrcd.PROTOCOL_SERVER_VBS_OBJECT_LIST)

	def vbs_query(self, machine, objid, extra=0):
		"""
		Query vbs data from a specific vbs object

			>>> rt.vbs_query('0010F3067C3A', 24)
			[{'bidirectional': 0L,
			  'bidirectional speed': 0,
			  'extra': -183572567,
			  'id': 24,
			  'inbound': 0L,
			  'inbound speed': 0,
			  'machineid': '0010F3067C3A',
			  'name': 'VBS Grund',
			  'outbound': 0L,
			  'outbound speed': 0,
			  'splitname': '169.231.14.245'}]

		@type machine: C{str}
		@param machine: MachineID of the PacketLogic you want to query.

		@type objid: C{int}
		@param objid: ID of VBS shapingobject

		@type extra: C{int}
		@param extra: Extra key; netobject id in case of split
		              by netobject, or ip number in case of
		              split by local host (encoded as an
		              integer). If argument isn't given ALL
		              available split keys are listed.

		@rtype: C{list} of C{dict}

		@returns: A list of dictionaries, where each
		          dictionary holds VBS data with the following
		          contents: 
		           - 'machineid', 'id', 'name'
		                identify the shaping object.
		           - 'inbound/outbound/bidirectional'
		                accumulated volume of transferred data
		           - 'inbound/outbound/bidirectional speed'
		                holds the current speed
		           - 'splitname'
		                text representation of the split instance of the
		                shaping object (for example for split by
		                local host, the local host IP address as a
		                string).
		           - 'extra' 
		                numerical representation of the split
		                instance of the shaping object (for
		                example for split by local host; an
		                integer representation of the IP
		                address, for split by netobject; the
		                id number of the netobject). This is
		                the number that would be passed as
		                'extra' argument to L{vbs_reset} or
		                L{vbs_query}.

		@raises PLDPermissionError: If the user lacks permission for requested action

		@attention: This function isn't supported in v12.0 and
		            v12.1 firmwares.
		@attention: In v12.2 firmware or newer the C{machineid}
		            argument is ignored.
		@attention: In v12.2 firmware or newer bidirectional
		            VBS isn't supported, therefore the
		            'bidirectional' value returned is always 0.
		"""
		PLrcd.vbsQuery(self.io, machine, objid, extra)
		lc = self._ListCollector(PLrcd.handleVbsQuery)
		self._handler_add(PLrcd.PROTOCOL_SERVER_VBS_REPLY, lc)
		# run until we get VBS_REPLY_END
		self._do(None, PLrcd.PROTOCOL_SERVER_VBS_REPLY_END)
		del self.handlers[PLrcd.PROTOCOL_SERVER_VBS_REPLY]
		return lc.lst


	def add_shapingcnt_callback(self, func):
		"""Add a callback for byte counters.

		The signature of the callback should be:

		C{cb(idx, oid, sk, flags, _in, _out)}
		
		idx is counter index in PL (i.e a number between 0 and MAX_COUNTERS)
		oidx is shaping object id
		sk is split key (i.e netobject id on split-by-netobject, ip on
		split-by-localhost, subscriber name on split-by-subscriber, and 0 on no
		split)
		flags is flags bitmask, if (1<<13) is set it means that this is the last
		update for this counter, and it will be reset to 0.
		_in and _out are bytecounters.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.2 firmware or newer.
		"""
		self.shapingcnt_callbacks.append(func)
		if len(self.shapingcnt_callbacks) > 1:
			return
		PLrcd.shapingcounterShow(self.io)
		self._flush()

	def shapingcnt_request_all(self):
		"""
		Requests all shaping counters, regardless if they have
		changed or not.
		
		@attention: This function requires v12.2 firmware or newer.
		"""
		PLrcd.shapingcounterRequestAll(self.io)
		self._flush()

	def add_update_done_callback(self, func):
		"""
		Add a callback to call after each update is done.

		@type func: C{callable}
		@param func: function to call. The function is called
		             without any arguments

		"""
		self.update_done_callbacks.append(func)

	def _update_once(self):
		""" Wait for a stats/channel callback and return."""
		last = time.time()
		self.updating = True
		self._update_once_data = None
		PLrcd.requestUpdate(self.io)
		timeout = self.timeout
		if timeout is None:
			timeout = sys.maxint
		while self.updating:
			if time.time() > last + timeout:
				raise IOError("Operation timed out")
			if self._process(5):
				last = time.time()
				
		return self._update_once_data

	def _update_once_callback(self, data):
		""" Callback for the _update_once function."""
		
		self._update_once_data = data
		self.stop_updating()
		
	def update_forever(self, interval=5):
		"""
		Start updating.

		@type interval: C{int}
		@param interval: Delay between updates.

		@raises PLDPermissionError: If the user lacks permission for any of the registerd callbacks

		"""

		next = time.time()
		last = time.time()
		self.updating = True
		timeout = self.timeout
		if timeout is None:
			timeout = sys.maxint
		while self.updating:
			now = time.time()
			if now > last + timeout:
				raise IOError("Operation timed out")
			if now > next:
				PLrcd.requestUpdate(self.io)
				next = now + interval
			if self._process(next - now):
				last = time.time()

	def stop_updating(self):
		"""
		Stop the updating and clear callback functions.

		"""
		self.updating = False
		self.shapingcnt_callbacks = []
		PLrcd.shapingcounterHide(self.io)

	def drdl_revision(self):
		"""
		Retrieve revision string for currently loaded drdl ARM.

		@returns: Revision string.
		@rtype: C{str}

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.1 firmware or newer.
		"""
		PLrcd.drdlRevision(self.io)
		return self._do(PLrcd.handleDrdlRevision, PLrcd.PROTOCOL_SERVER_DRDL_REVISION)

	def reload_drdl(self):
		"""
		Reload drdl signatures. For internal use only.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.1 firmware or newer.
		"""

		PLrcd.drdlReload(self.io)
		self._fullflush()

	def wait_for_ruleset(self, commitid, timeout=60):
		"""
		Waits for a ruleset of 'commitid' to be loaded and compiled by plrcd.
		wait_for_ruleset() will return when plrcd sees a commitid that is
		atleast as new as 'commitid' given. If plrcd already has loaded
		or loads a newer commitid than the given 'commitid' it will also return
		with True.

		Typical use:
			>>> # Add Test netobject to pldb
			>>> test = rs.object_add("/NetObjects/Test")
			>>> commitid = rs.commit()
			>>> # Wait for Test netobject to be visible in plrcd
			>>> rt.wait_for_ruleset(commitid)
			>>> # Add a dynnetobj item in plrcd under Test netobject
			>>> rt.dyn.add(test.id, "10.0.0.1")

		@type commitid: C{CommitID}
		@param commitid: CommitID to wait for.

		@type timeout: C{int}
		@param timeout: Max seconds to wait before returning with False.

		@returns: True if commitid has been seen. False if timeout occured.
		@rtype: C{boolean}

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v14.0 firmware or newer.
		"""

		start = time.time()
		if timeout is None:
			timeout = sys.maxint
		while self.commit_bcast < commitid:
			now = time.time()
			if now > start + timeout:
				raise IOError("wait_for_ruleset() timed out after %d seconds" % timeout);
			self._process(timeout - (now - start))
		return self.commit_bcast

	def current_ruleset(self):
		"""
		Returns current ruleset's commitid. This also work directly after
		connecting where no previous rs.commit() has been performed.

		@returns: Commitid of current plrcd loaded ruleset
		@rtype commitid: C{CommitID}
		"""
		# If we just connected, wait for the first commitid to be processed
		while self.commit_bcast is None:
			self._process(1)
		return self.commit_bcast
	current_ruleset = property(current_ruleset)

	# Make an instance of KeyVault reachable
	_keyvault_doc = """
		Get a handle for KeyVault operations. See L{KeyVault} object for details.
"""
	keyvault = property(lambda self: KeyVault(self), doc=_keyvault_doc)
