"""
This module contains functionality for accessing Host triggers.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> ht = pl.HostTriggers()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from resource import Resource
from thisversion import PLDB

import plutil
from packetlogic2.exceptions import PLDBUnsupportedInFirmware

_Prop_HostTrigger = plutil.Prop('id', 'name', 'script', 'in_bw', 'out_bw',
                                'in_fwd', 'out_fwd', 'in_cps', 'out_cps',
                                'conns', 'uconns', 'uconn_rate', 'out_ttls',
                                'qoe_in_int', 'qoe_in_ext', 'qoe_out_int', 'qoe_out_ext',
                                'netobject', 'trigtype')

class HostTrigger(object):
    """
    Information about a Host trigger.

    Returned in a list by L{HostTriggers.hosttrigger_list}.
    """

    __slots__ = ('_prop', '_pldb')

    def __init__(self, _pldb=None, **props):
        """
        Should not be instantiated manually, use L{hosttrigger_list <HostTriggers.hosttrigger_list>}.
        """
        self._pldb = _pldb
        self._prop = _Prop_HostTrigger(**props)

    def __repr__(self):
        return "<%s #%d '%s'>" % (self.__class__.__name__, self._prop.id, self._prop.name)

    def update(self, script=None, in_bw=None, out_bw=None, in_cps=None, out_cps=None,
               conns=None, uconns=None, uconn_rate=None, out_ttls=None,
               qoe_in_int=None, qoe_in_ext=None, qoe_out_int=None, qoe_out_ext=None,
               netobject=None, trigtype=None):
        """
        Updates the host trigger properties.

        @type script: C{str}
        @param script: The script filename to execute

        @type in_bw: C{int}
        @param in_bw: in bandwidth limit to trigger on in Kbps

        @type out_bw: C{int}
        @param out_bw: out bandwidth limit to trigger on in Kbps

        @type in_cps: C{int}
        @param in_cps: in connections per second to trigger on

        @type out_cps: C{int}
        @param out_cps: out connections per second to trigger on

        @type conns: C{int}
        @param conns: no of concurrent connections to trigger on

        @type uconns: C{int}
        @param uconns: no of unest. connections to trigger on

        @type uconn_rate: C{int}
        @param uconn_rate: unest. connection rate to trigger on

        @type out_ttls: C{int}
        @param out_ttls: out number of different used ttls to trigger on

        @type qoe_in_int: C{int}
        @param qoe_in_int: QoE for in traffic on internal interface

        @type qoe_in_ext: C{int}
        @param qoe_in_ext: QoE for in traffic on external interface

        @type qoe_out_int: C{int}
        @param qoe_out_int: QoE for out traffic on internal interface

        @type qoe_out_ext: C{int}
        @param qoe_out_ext: QoE for out traffic on external interface

        @type netobject: C{int}
        @param netobject: Netobject id the trigger is configured for when trigtype in (TRIGTYPE_HOST_IN_NETOBJECT, TRIGTYPE_HOST_IN_NETOBJECT, TRIGTYPE_HOST_IN_NETOBJECT)

        @type trigtype: C{int}
        @param trigtype: Trigger type. One of TRIGTYPE_HOST, TRIGTYPE_HOST_IN_NETOBJECT, TRIGTYPE_HOST_IN_NETOBJECT, TRIGTYPE_HOST_IN_NETOBJECT

        @raises PLDBUnsupportedInFirmware: If this method is called without firmware support.

        @attention: This method requires firmware v14.1 or newer
        """
        if script is None:
            script = self._prop.script
        if in_bw is None:
            in_bw = self._prop.in_bw
        if out_bw is None:
            out_bw = self._prop.out_bw
        if in_cps is None:
            in_cps = self._prop.in_cps
        if out_cps is None:
            out_cps = self._prop.out_cps
        if conns is None:
            conns = self._prop.conns
        if uconns is None:
            uconns = self._prop.uconns
        if uconn_rate is None:
            uconn_rate = self._prop.uconn_rate
        if out_ttls is None:
            out_ttls = self._prop.out_ttls
        if qoe_in_int is None:
            qoe_in_int = self._prop.qoe_in_int
        if qoe_in_ext is None:
            qoe_in_ext = self._prop.qoe_in_ext
        if qoe_out_int is None:
            qoe_out_int = self._prop.qoe_out_int
        if qoe_out_ext is None:
            qoe_out_ext = self._prop.qoe_out_ext
        if netobject is None:
            netobject = self._prop.netobject
        if trigtype is None:
            trigtype = self._prop.trigtype

        plutil._do(self._pldb,
                   self._pldb.session.hosttrigger_update(self._prop.id, script, in_bw, out_bw,
                                                         in_cps, out_cps, conns, uconns, uconn_rate, out_ttls,
                                                         qoe_in_int, qoe_in_ext, qoe_out_int, qoe_out_ext,
                                                         netobject, trigtype),
                   plutil._AddCB(self._pldb))

        self._prop.script =script
        self._prop.in_bw = in_bw
        self._prop.out_bw = out_bw
        self._prop.in_cps = in_cps
        self._prop.out_cps = out_cps
        self._prop.conns = conns
        self._prop.uconns = uconns
        self._prop.uconn_rate = uconn_rate
        self._prop.out_ttls = out_ttls
        self._prop.qoe_in_int = qoe_in_int
        self._prop.qoe_in_ext = qoe_in_ext
        self._prop.qoe_out_int = qoe_out_int
        self._prop.qoe_out_ext = qoe_out_ext
        self._prop.netobject = netobject
        self._prop.trigtype = trigtype

    def remove(self):
        """
        Remove this host trigger from the system.
        """
        plutil._do(self._pldb, self._pldb.session.hosttrigger_remove(self._prop.id), plutil._AddCB(self._pldb))

    id = property(lambda self: self._prop.id, doc="ID Number of the host trigger")
    name = property(lambda self: self._prop.name, doc="Name of the host trigger")
    script = property(lambda self: self._prop.script, doc="Script filename to execute")
    in_bw = property(lambda self: self._prop.in_bw, doc="Inbound bandwidth limit to trigger on in Kbps")
    out_bw = property(lambda self: self._prop.out_bw, doc="Outbound bandwidth limit to trigger on in Kbps")
    in_fwd = property(lambda self: 0, doc="Inbound forwarded bandwidth limit to trigger on in Kbps. (Deprecated and unused in firmware version v12.0 and newer)")
    out_fwd = property(lambda self: 0, doc="Outbound forwarded bandwidth limit to trigger on in Kbps. (Deprecated and unused in firmware version v12.0 and newer)")
    in_cps = property(lambda self: self._prop.in_cps, doc="Inbound connections per second to trigger on")
    out_cps = property(lambda self: self._prop.out_cps, doc="Outbound connections per second to trigger on")
    conns = property(lambda self: self._prop.conns, doc="Number of concurrent connections to trigger on")
    uconns = property(lambda self: self._prop.uconns, doc="Number of concurrent unest. connections to trigger on")
    uconn_rate = property(lambda self: self._prop.uconn_rate, doc="Unest. connection rate to trigger on")
    out_ttls = property(lambda self: self._prop.out_ttls, doc="Number of different TTL values seen on outbound traffic to trigger on")
    qoe_in_int = property(lambda self: self._prop.qoe_in_int, doc="QoE for inbound traffic on internal interface to trigger on. (Requires firmware version v14.1 or newer)")
    qoe_in_ext = property(lambda self: self._prop.qoe_in_ext, doc="QoE for inbound traffic on external interface to trigger on. (Requires firmware version v14.1 or newer)")
    qoe_out_int = property(lambda self: self._prop.qoe_out_int, doc="QoE for outbound traffic on internal interface to trigger on. (Requires firmware version v14.1 or newer)")
    qoe_out_ext = property(lambda self: self._prop.qoe_out_ext, doc="QoE for outbound traffic on external interface to trigger on. (Requires firmware version v14.1 or newer)")
    netobject = property(lambda self: self._prop.netobject, doc="Netobject id the trigger is configured for when trigtype in (TRIGTYPE_HOST_IN_NETOBJECT, TRIGTYPE_HOST_IN_NETOBJECT, TRIGTYPE_HOST_IN_NETOBJECT) (Requires firmware version v14.1 or newer)")
    trigtype = property(lambda self: self._prop.trigtype, doc="Trigger type. One of TRIGTYPE_HOST, TRIGTYPE_HOST_IN_NETOBJECT, TRIGTYPE_HOST_IN_NETOBJECT, TRIGTYPE_HOST_IN_NETOBJECT (Requires firmware version v14.1.2 or newer)")

_NONEXISTANT_TRIGTYPE = object()

class HostTriggers(Resource):
    def _resource(self):
        return PLDB.RESOURCE_HOSTTRIGGER

    def list(self):
        """
        List all host triggers in the PacketLogic system.

            >>> ht.hosttrigger_list()
            [<HostTrigger #1 'filter hosts with 100 CPS'>]

        @rtype: C{list} of L{HostTrigger}
        @returns: List containing L{HostTrigger} objects.
        """
        return plutil._do(self, self.session.hosttrigger_list(),
                       plutil._ListCB(self, self.session.hosttrigger_list_handle, HostTrigger))

    #enum for trigtype
    TRIGTYPE_HOST = PLDB.HOSTTRIGGER_TRIGTYPE_HOST
    TRIGTYPE_HOST_IN_NETOBJECT = PLDB.HOSTTRIGGER_TRIGTYPE_HOST_IN_NETOBJECT
    TRIGTYPE_NETOBJECT = PLDB.HOSTTRIGGER_TRIGTYPE_NETOBJECT
    TRIGTYPE_NETOBJECT_CHILD = PLDB.HOSTTRIGGER_TRIGTYPE_NETOBJECT_CHILD

    def add(self, name, script, in_bw=0, out_bw=0, in_fwd=0, out_fwd=0,
            in_cps=0, out_cps=0, conns=0, uconns=0, out_ttls=0,
            qoe_in_int=0, qoe_in_ext=0, qoe_out_int=0, qoe_out_ext=0,
            netobject=0, trigtype=TRIGTYPE_HOST, uconn_rate=0):
        """
        Adds a new host trigger to the system.

        QoE (Quality of Experience) triggers can have values from 100 to 0
        (Corresponding to 100% to 0%). It will trigger when QoE goes below
        configured value. See product guide for a detailed definition
        of QoE values.

        @type name: C{str}
        @param name: The new triggers name

        @type script: C{str}
        @param script: The script filename to execute

        @type in_bw: C{int}
        @param in_bw: Inbound bandwidth limit to trigger on in Kbps

        @type out_bw: C{int}
        @param out_bw: Outbound bandwidth limit to trigger on in Kbps

        @type in_fwd: C{int}
        @param in_fwd: Inbound forwarded bandwidth limit to trigger on in Kbps (not available in firmware versions v12.0 and newer)

        @type out_fwd: C{int}
        @param out_fwd: Outbound forwarded bandwidth limit to trigger on in Kbps (not available in firmware versions v12.0 and newer)

        @type in_cps: C{int}
        @param in_cps: Inbound connections per second to trigger on

        @type out_cps: C{int}
        @param out_cps: Outbound connections per second to trigger on

        @type conns: C{int}
        @param conns: Number of concurrent connections to trigger on

        @type uconns: C{int}
        @param uconns: Number of concurrent unestablished connections to trigger on

        @type out_ttls: C{int}
        @param out_ttls: Number of different TTL values seen on outbound traffic to trigger on (requires firmware 13.1 or newer)

        @type qoe_in_int: C{int}
        @param qoe_in_int: QoE for in traffic on internal interface. (requires firmware 14.1 or newer)

        @type qoe_in_ext: C{int}
        @param qoe_in_ext: QoE for in traffic on external interface. (requires firmware 14.1 or newer)

        @type qoe_out_int: C{int}
        @param qoe_out_int: QoE for out traffic on internal interface. (requires firmware 14.1 or newer)

        @type qoe_out_ext: C{int}
        @param qoe_out_ext: QoE for out traffic on external interface. (requires firmware 14.1 or newer)

        @type netobject: C{int}
        @param netobject: Netobject id the trigger is configured for when trigtype in (TRIGTYPE_HOST_IN_NETOBJECT, TRIGTYPE_NETOBJECT or TRIGTYPE_NETOBJECT_CHILD)

        @type trigtype: C{int}
        @param trigtype: Trigger type. One of TRIGTYPE_HOST, TRIGTYPE_HOST_IN_NETOBJECT, TRIGTYPE_NETOBJECT or TRIGTYPE_NETOBJECT_CHILD

        @type uconn_rate: C{int}
        @param uconn_rate: Unestablished connection rate to trigger on. (requires firmware 15.0 or newer)

        @attention: in_fwd and out_fwd are deprecated and will generate an exception if they have other values than 0.
        @attention: Triggering on out_ttls requires v13.1 firmware or newer.
        @attention: Triggering on qoe values requires v14.1 firmware or newer.
        @attention: Trigger types TRIGTYPE_HOST_IN_NETOBJECT, TRIGTYPE_NETOBJECT or TRIGTYPE_NETOBJECT_CHILD requires 14.1 firmware or newer.
        @attention: Triggering on uconn_rate requires v15.0 firmware or newer.
        """
        if in_fwd != 0:
            raise PLDBUnsupportedInFirmware("Usage of 'in_fwd' is deprecated in this firmware version. 'in_fwd' must be set to 0.")
        if out_fwd != 0:
            raise PLDBUnsupportedInFirmware("Usage of 'out_fwd' is deprecated in this firmware version. 'out_fwd' must be set to 0.")

        if trigtype is _NONEXISTANT_TRIGTYPE:
            raise PLDBUnsupportedInFirmware("Unsupported trigger type")


        _id = plutil._do(self, self.session.hosttrigger_add(name, script, in_bw, out_bw, in_cps, out_cps,
                                                            conns, uconns, uconn_rate, out_ttls,
                                                            qoe_in_int, qoe_in_ext, qoe_out_int, qoe_out_ext,
                                                            netobject, trigtype),
                         plutil._AddCB(self, self.session.hosttrigger_add_handle))
        return HostTrigger(self, id=_id, name=name, script=script,
                           in_bw=in_bw, out_bw=out_bw,
                           in_cps=in_cps, out_cps=out_cps,
                           conns=conns, uconns=uconns, uconn_rate=uconn_rate,
                           out_ttls=out_ttls,
                           qoe_in_int=qoe_in_int, qoe_in_ext=qoe_in_ext,
                           qoe_out_int=qoe_out_int, qoe_out_ext=qoe_out_ext,
                           netobject=netobject, trigtype=trigtype)

    hosttrigger_list = plutil.deprecatedalias("hosttrigger_list", "list")
    hosttrigger_add = plutil.deprecatedalias("hosttrigger_add", "add")
